import argparse
import json
import re
import google
import subprocess
from google.cloud import storage

from winval.workflow_var import WorkflowVarType
from winval.wdl_parser import WdlParser
from winval import logger


def get_args():
    parser = argparse.ArgumentParser('cloud_files_validator', description='Validate GCS inputs actually exist')
    parser.add_argument('--json', required=True)
    parser.add_argument('--wdl', required=True)
    return parser.parse_args()


def split_uri(uri: str):
    matches = re.match(r"gs://(.*?)/(.*)", uri)
    if matches:
        return matches.groups()
    else:
        return None, None


def blob_exists(bucket_name, filepath):
    client = storage.Client()
    bucket = client.get_bucket(bucket_name)
    blob = bucket.blob(filepath)
    return blob.exists()


class CloudFilesValidator:

    def __init__(self, wdl_file: str, json_file: str):
        logger.debug('-----------------------------')
        logger.debug('---- CloudFilesValidator ----')
        logger.debug('-----------------------------')
        self.json_file = json_file
        self.workflow_vars = None
        wdl_inputs = WdlParser(wdl_file).parse_workflow_variables()
        self.workflow_vars = wdl_inputs.workflow_vars
        with open(json_file) as jf:
            wdl_inputs.fill_values_from_json(json.load(jf))
        self.validated_files = []
        self.non_validated_files = []

    def validate(self) -> bool:
        for workflow_var in self.workflow_vars.values():
            if workflow_var.type == WorkflowVarType.FILE and workflow_var.value is not None:
                self.validate_file(workflow_var.value)
            elif workflow_var.type == WorkflowVarType.FILE_ARRAY and workflow_var.value is not None:
                for file_uri in workflow_var.value:
                    self.validate_file(file_uri)

        logger.debug('-------------------------------------------')
        summary_str = f"Existing URI's: {len(self.validated_files)}, Non-existing URI's: {len(self.non_validated_files)}"
        logger.info(summary_str)
        for fn in self.non_validated_files:
            logger.error(f'Missing URI: {fn}')

        return len(self.non_validated_files) == 0

    def validate_file(self, file_uri: str):
        file_uri = file_uri.replace('"', '')
        if 'gs://' not in file_uri:
            return
        bucket_name, file_name = split_uri(file_uri)
        if bucket_name is None:
            self.non_validated_files.append(file_uri)
        try:
            if blob_exists(bucket_name, file_name):
                self.validated_files.append(file_uri)
            else:
                self.non_validated_files.append(file_uri)
        # blob_exists requires storage.bucket.get access which might be forbidden, 
        # yet objects in this bucket can still be accessed by gsutil
        # Could not fine a python api to access the object in this case
        except google.api_core.exceptions.Forbidden:
            process = subprocess.run(f'gsutil ls {file_uri}'.split(), capture_output=True)
            if process.returncode == 0:
                self.validated_files.append(file_uri)
            else:
                self.non_validated_files.append(file_uri)

    def validate_files_in_object(self, o):
        if type(o) == dict:
            for key, value in o.items():
                self.validate_files_in_object(value)
        elif type(o) == list:
            for value in o:
                self.validate_files_in_object(value)
        elif type(o) == str:
            self.validate_file(o)


def main():
    """
    validate all GCP URLs in json inputs point to existing objects
    """
    args = get_args()
    validated = CloudFilesValidator(args.wdl, args.json).validate()
    if not validated:
        raise RuntimeError('Found non existing files!')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['winval', 'winval.antlr']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime==4.10',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'pytest>=7.1.2,<8.0.0']

setup_kwargs = {
    'name': 'winval',
    'version': '0.5.1',
    'description': 'WDL workflow inputs validation',
    'long_description': '# winval\n### Workflow inputs validation python library\n\n* Currently, supports WDL workflows.\n* Constraints on inputs can be written as specially tagged (#@wv) comments in the WDL inputs section.\n* The syntax of constraints is a python-similar DSL, with ANTLR defined grammar.\n* The WDL + inputs.json can be validated using a script/function before submitting a workflow.\n\n## Usage:\n\n### validate_wdl_constraints\nFrom python\n~~~\nfrom winval.validate_wdl_constraints import run_winval\nis_validated = run_winval(wdl_file, json_file)\n~~~\n\nFrom unix command-line:\n~~~\npython winval/validate_wdl_constraints.py --wdl <wdl_file> --json <json_file>\n~~~\n\n### cloud_files_validater\n#### Make sure google-storage permissions are equivalent to the batch workers permissions \n\nFrom python\n~~~\nfrom winval.cloud_files_validator import CloudFilesValidator\nis_validated = CloudFilesValidator(args.wdl, args.json).validate()\n~~~\n\nFrom unix command-line:\n~~~\npython winval/cloud_files_validator.py --wdl <wdl_file> --json <json_file>\n~~~\n\n## WDL constraints example\n~~~\nworkflow MyWorkflow {\n\n  input {\n    File file\n    Int c\n    File* opt_file_1\n    File* opt_file_2\n    Array[File] files\n    Array[File] index_files\n    MyStruct struct_instance\n\n    #@wv defined(opt_file_1) <-> defined(opt_file_2)\n    #@wv defined(opt_file_1) -> c > 1\n    #@wv len(files) == len(index_files)\n    #@wv len(files) >= 0\n    #@wv len(index_files) >= 0\n    #@wv c <= 1 and c >= 0\n    #@wv suffix(file) == ".fasta"\n    #@wv suffix(files) <= {".bam", ".cram"} \n    #@wv prefix(index_files) == files\n    #@wv len(struct_instance[\'field_a\']) > 0\n  }\n  ...\n}\n  \n  struct MyStruct{\n     String field_a,\n     String field_b\n  }\n~~~\n\n## Generate parsers from grammar:\n~~~\ncd <project_root>/winval\nantlr4 -Dlanguage=Python3 winval.g4 -visitor -o antlr\n~~~\n\n## Available atomic expressions:\n* int: 5\n* float: 5.6\n* bool: True, False\n* str: "some string", \'some_string\'\n* workflow_variable: my_var\n* evaluates to value given by json conf, or None if not defined in json\n* empty_set: {} \n\n## Available python operators\n* `+`,`-`,`*`,`**`,`/`,`&`,`|`,`%`\n* `and`, `or`, `in`\n* `<`, `<=`, `==`, `>=`, `>`, `!=`\n* Notice the following useful operators work for sets:\n  * `-`: set subtraction\n  * `&`: set intersection\n  * `|` : is set union \n  * `<=`:  subset \n\n## Available python functions\n* `len()`, `not()`\n* `basename()`, `splitext()` (from os.path)\n\n## Available convenience functions and operators:\n* `x <-> y`: if-and-only-if logical operator (if x is true if and only if y is true)\n* `x -> y`: implies logical operator (if x is true then y should be true)\n* `defined(x)`: if x a defined variable\n* `prefix(x)`: return path prefix of String/File or list of prefixes for \n* `suffix(x)`: return path suffix of String/File or set of suffixes for Array[String]/Array[File]\n',
    'author': 'doron',
    'author_email': 'doron.shemtov@ultimagen.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
